"""Visualize notes stored in database."""

import sqlite3
import sys

class MultiDiGraph:
    """Represents directed graph with parallel edges."""
    def __init__(self):
        self.attributes = {}

    @property
    def edges(self):
        """Generate edges with attributes."""
        for edge, attrs in self.attributes.items():
            src, dest = edge
            for attr in attrs:
                yield src, dest, attr

    def add_edge(self, src, dest, **kwargs):
        """Insert edge from src to dest with attributes in kwargs."""
        self.attributes.setdefault((src, dest), [])
        self.attributes[(src, dest)].append(kwargs.copy())

    def to_dot(self):
        """Generate dot graph."""
        yield "digraph {\n"
        yield "\trankdir=LR;\n"
        for src, dest, attrs in self.edges:
            yield "\t{} -> {}{};\n".format(src, dest, "" if not attrs else\
                    "[{}]".format(",".join(f"{k}={v}" for k, v in attrs.items())))
        yield "}\n"

def fetch_link_graph(con):
    """Get links from database and return a MultiDiGraph."""
    sql = """
        SELECT src, dest, S.title, D.title
            FROM Links JOIN Notes AS S ON src = S.id
                JOIN Notes AS D ON dest = D.id
    """
    graph = MultiDiGraph()
    for src, dest, src_title, dest_title in con.execute(sql):
        graph.add_edge(f'"[{src}] {src_title}"', f'"[{dest}] {dest_title}"')
    return graph

def fetch_sequence_graph(con):
    """Get sequences from database and return a MultiDiGraph."""
    sql = """
        SELECT NP.id, NN.id, NP.title, NN.title
            FROM Sequences JOIN Aliases AS AP ON prev = AP.alias
                JOIN Aliases AS AN ON next = AN.alias
                    JOIN Notes AS NP ON AP.id = NP.id
                        JOIN Notes AS NN ON AN.id = NN.id
    """
    graph = MultiDiGraph()
    for parent, child, ptitle, ctitle in con.execute(sql):
        graph.add_edge(f'"[{parent}] {ptitle}"', f'"[{child}] {ctitle}"')
    return graph

def fetch_backlinks(con):
    """Get backlinks from database and return a MultiDiGraph."""
    sql = """
        SELECT dest, D.title, src, S.title FROM StrongLinks
            JOIN Notes AS D ON D.id = dest
                JOIN Notes AS S ON S.id = src
    """
    graph = MultiDiGraph()
    for dest, dtitle, src, stitle in con.execute(sql):
        graph.add_edge(f'"[{dest}] {dtitle}"', f'"[{src}] {stitle}"')
    return graph

def write_dot_graph(con, direct=False, sequence=False, backlinks=False,
                    out=sys.stdout):
    """Generate dot graph from database."""
    graph = MultiDiGraph()
    if direct:
        for src, dest, _ in fetch_link_graph(con).edges:
            graph.add_edge(src, dest)
    if sequence:
        for src, dest, _ in fetch_sequence_graph(con).edges:
            graph.add_edge(src, dest, color="red")
    if backlinks:
        for src, dest, _ in fetch_backlinks(con).edges:
            graph.add_edge(src, dest, style="dashed")
    if isinstance(out, str):
        out = open(out, "w")
    with out:
        for line in graph.to_dot():
            print(line, file=out, end="")

if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser(
        description="Generate a dot file to visualize your slipbox notes.")
    parser.add_argument("database", help="filename of sqlite3 database")
    parser.add_argument("-d", "--direct", action="store_true",
                        help="show direct links")
    parser.add_argument("-s", "--sequence", action="store_true",
                        help="show sequence links")
    parser.add_argument("-b", "--backlink", action="store_true",
                        help="show backlinks")
    parser.add_argument("-o", "--out", default=sys.stdout, help="output file")
    args = parser.parse_args()
    with sqlite3.connect(args.database) as conn:
        write_dot_graph(conn, args.direct, args.sequence, args.backlink,
                        args.out)



"""Test graph.py."""

from .graph import fetch_link_graph, fetch_sequence_graph
from .graph import fetch_backlinks, MultiDiGraph, write_dot_graph
from .mock import mock_database
from .utils import make_temporary_file

SAMPLE_SCRIPT_WITH_PARALLEL_EDGES = """
    PRAGMA foreign_keys=ON;
    INSERT INTO Files (filename) VALUES ('test.md');
    INSERT INTO Notes (id, title, filename) VALUES
        (0, 'root', 'test.md'),
        (1, 'src', 'test.md'),
        (2, 'dest', 'test.md');
    INSERT INTO Links (src, dest, annotation) VALUES
        (1, 2, 'annotation');
    INSERT INTO Aliases (id, owner, alias) VALUES
        (1, 0, '0a'),
        (2, 0, '0a1');
    INSERT INTO Sequences (prev, next) VALUES
        ('0a', '0a1');
"""

def sample_script():
    """Sample script to initialize mock database."""
    return """
    PRAGMA foreign_keys=ON;
    INSERT INTO Files (filename) VALUES ('test.md');
    INSERT INTO Notes (id, title, filename) VALUES
        (1, 'Note 1', 'test.md'),
        (2, 'Note 2', 'test.md'),
        (3, 'Note 3', 'test.md'),
        (4, 'Note 4', 'test.md'),
        (5, 'Note 5', 'test.md'),
        (6, 'Note 6', 'test.md'),
        (7, 'Note 7', 'test.md'),
        (8, 'Note 8', 'test.md'),
        (9, 'Note 9', 'test.md');
    -- i -> 2*i
    INSERT INTO Links (src, dest, annotation) VALUES
        (1, 2, 'annotation'),
        (2, 4, 'annotation'),
        (3, 6, ''),
        (4, 8, '');
    INSERT INTO Aliases (id, owner, alias) VALUES
        (2, 1, '1a'),
        (3, 1, '1a1'),
        (4, 1, '1a2'),
        (6, 5, '5a'),
        (7, 5, '5a1'),
        (8, 5, '5a2'),
        (9, 5, '5a3');
    INSERT INTO Sequences (prev, next) VALUES
        ('1a', '1a1'),
        ('1a', '1a2'),
        ('5a', '5a1'),
        ('5a', '5a2'),
        ('5a', '5a3');
    """

def sample_sequence_edge(src, dest):
    """Create edge with labeled nodes based on the sample data."""
    name = lambda i: f'"[{i}] Note {i}"'
    return (name(src), name(dest))

def test_fetch_link_graph():
    """Graph must contain exactly the edges that appear in the Links table,
    including unannotated links.
    """
    with mock_database(sample_script()) as conn:
        graph = fetch_link_graph(conn)
    edges = list(graph.edges)
    assert len(edges) == 4

    pairs = [(a, b) for a, b, _ in edges]
    for i in range(1, 5):
        assert sample_sequence_edge(i, 2*i) in pairs
    assert sample_sequence_edge(1, 3) not in pairs

def test_fetch_sequence_graph():
    """Sequence graph must contain exactly the edges that appear in the
    Sequences table.
    """
    with mock_database(sample_script()) as conn:
        graph = fetch_sequence_graph(conn)
    edges = list(graph.edges)
    assert len(edges) == 5

    pairs = [(a, b) for a, b, _ in edges]
    assert sample_sequence_edge(2, 3) in pairs
    assert sample_sequence_edge(2, 4) in pairs
    assert sample_sequence_edge(6, 7) in pairs
    assert sample_sequence_edge(6, 8) in pairs
    assert sample_sequence_edge(6, 9) in pairs

def test_fetch_backlinks():
    """Backlinks are reversed links with annotations."""
    with mock_database(sample_script()) as conn:
        graph = fetch_backlinks(conn)
    edges = list(graph.edges)
    assert len(edges) == 2

    pairs = [(a, b) for a, b, _ in edges]
    assert sample_sequence_edge(2, 1) in pairs
    assert sample_sequence_edge(4, 2) in pairs

def test_multidigraph_edd_edge():
    """MultiDiGraph.add_edge must insert endpoints and attributes."""
    graph = MultiDiGraph()
    assert not list(graph.edges)
    graph.add_edge(1, 2, color="red")
    graph.add_edge(2, 3)

    edges = list(graph.edges)
    pairs = [(a, b) for a, b, _ in edges]
    assert (1, 2) in pairs
    assert (2, 3) in pairs
    assert (1, 2, {"color": "red"}) in edges
    assert (2, 3, {}) in edges

def test_multidigraph_edges():
    """It must be possible to iterate through edges multiple times."""
    graph = MultiDiGraph()
    graph.add_edge(0, 0)
    graph.add_edge(0, 0)

    count = 0
    for _ in graph.edges:
        count += 1
    assert count == 2
    count = 0
    for _ in graph.edges:
        count += 1
    assert count == 2

def test_multidigraph_add_parallel_edges():
    """MultiDiGraph must be able to store parallel edges between two nodes."""
    graph = MultiDiGraph()
    graph.add_edge(1, 2)
    graph.add_edge(1, 2, color="red")
    graph.add_edge(1, 2, style="dashed")
    edges = list(graph.edges)
    assert len(edges) == 3
    attrs = []
    for src, dest, attr in edges:
        assert (src, dest) == (1, 2)
        attrs.append(attr)
    assert len(attrs) == 3
    assert attrs[0] == {}
    assert attrs[1] == {"color": "red"}
    assert attrs[2] == {"style": "dashed"}

def test_multidigraph_to_dot():
    """Test dot output."""
    graph = MultiDiGraph()
    graph.add_edge(1, 1, color="red")
    dot = graph.to_dot()
    assert "digraph {" in next(dot)
    assert "rankdir=LR;" in next(dot)
    line = next(dot)
    assert "1 -> 1" in line
    assert "color=red" in line
    assert "}" in next(dot)

def test_write_dot_graph():
    """Output dot graph must work with parallel edges."""
    with make_temporary_file() as temp,\
            mock_database(SAMPLE_SCRIPT_WITH_PARALLEL_EDGES) as conn:
        write_dot_graph(conn, True, True, True, temp)
        with open(temp) as file:
            lines = file.read().splitlines()

    dot = (line for line in lines)
    assert "digraph {" in next(dot)
    assert "rankdir=LR;" in next(dot)
    queue = sorted([next(dot), next(dot), next(dot)])
    assert "}" in next(dot)

    assert '"[1] src" -> "[2] dest";' in queue[0]
    assert '"[1] src" -> "[2] dest"' in queue[1]
    assert "color=red" in queue[1]
    assert '"[2] dest" -> "[1] src"' in queue[2]
    assert "style=dashed" in queue[2]
